ODA({ is: 'oda-ace-editor', imports: './src/ace.js',
    template: `
        <style>
            :host {
                display: block !important;
                position: relative;
                height: 100%;
                overflow: auto;
            }
            .ace_hidden-cursors { 
                opacity: {{showCursor ? 1 : 0}};
            }
        </style>
        <div @keydown.stop style="min-height: 100%;"></div>
    `,
    $public: {
        value: {
            $def: '',
            get() { 
                return this.editor?.getValue() || ''; 
            },
            set(n) { 
                this.setValue(n);
            }
        },
        theme: {
            $def: 'chrome',
            async set(n) {
                if (n)
                    await this.setTheme(n);
            },
            $list: ['ambiance', 'chaos', 'chrome', 'clouds', 'clouds_midnight', 'cobalt', 'crimson_editor', 'dawn', 'dracula', 'dreamweaver', 'eclipse', 'github', 'gob', 'gruvbox', 'idle_fingers', 'iplastic', 'katzenmilch', 'kr_theme', 'kuroir', 'merbivore', 'merbivore_soft', 'monokai', 'mono_industrial', 'pastel_on_dark', 'solarized_dark', 'solarized_light', 'sqlserver', 'tomorrow_night_bright', 'tomorrow_night_eighties', 'twilight', 'vibrant_ink', 'xcode']
        },
        mode: {
            $def: 'javascript',
            async set(n) {
                if (n)
                    await this.setMode(n);
            },
            $list: ['abap', 'abc', 'actionscript', 'ada', 'apache_conf', 'apex', 'applescript', 'aql', 'asciidoc', 'asl', 'assembly_x86', 'autohotkey', 'batchfile', 'bro', 'c9search', 'cirru', 'clojure', 'cobol', 'coffee', 'coldfusion', 'crystal', 'csharp', 'c_cpp', 'd', 'dart', 'diff', 'django', 'dockerfile', 'dot', 'drools', 'edifact', 'eiffel', 'ejs', 'elixir', 'elm', 'erlang', 'forth', 'fortran', 'fsl', 'ftl', 'gcode', 'gherkin', 'gitignore', 'glsl', 'gobstones', 'golang', 'graphqlschema', 'groovy', 'haml', 'handlebars', 'haskell', 'haskell_cabal', 'haxe', 'hjson', 'html', 'html_elixir', 'html_ruby', 'ini', 'io', 'jack', 'jade', 'java', 'javascript', 'json', 'json5', 'jsoniq', 'jsp', 'jssm', 'jsx', 'julia', 'kotlin', 'latex', 'less', 'liquid', 'lisp', 'livescript', 'logtalk', 'live_script', 'logiql', 'lsl', 'lua', 'luapage', 'lucene', 'makefile', 'markdown', 'mask', 'matlab', 'maze', 'mel', 'mixal', 'mushcode', 'mysql', 'nix', 'nsis', 'objectivec', 'ocaml', 'pascal', 'perl', 'perl6', 'pgsql', 'php', 'php_laravel_blade', 'pig', 'plain_text', 'powershell', 'praat', 'prolog', 'properties', 'protobuf', 'puppet', 'python', 'r', 'razor', 'rdoc', 'red', 'redshift', 'rhtml', 'rst', 'ruby', 'rust', 'sass', 'scad', 'scala', 'scheme', 'scss', 'sh', 'sjs', 'slim', 'smarty', 'snippets', 'soy_template', 'space', 'sparql', 'sql', 'sqlserver', 'stylus', 'svg', 'swift', 'swig', 'tcl', 'tex', 'text', 'textile', 'toml', 'tsx', 'turtle', 'twig', 'typescript', 'vala', 'vbscript', 'velocity', 'verilog', 'vhdl', 'visualforce', 'wollok', 'xml', 'xquery', 'yaml', 'zeek']
        },
        readOnly: {
            $def: false,
            set(n) {
                this.editor?.setReadOnly(n);
            }
        },
        fontSize: { $def: 16, set(n) { this.editor?.setOption('fontSize', n) } },
        wrap: { $def: true, set(n) { this.editor?.setOption('wrap', n) } },
        minLines: { $def: 1, set(n) { this.editor?.setOption('minLines', n) } },
        maxLines: { $def: '', set(n) { this.editor?.setOption('maxLines', n) } },
        showGutter: { $def: true, set(n) { this.editor?.setOption('showGutter', n) } },
        highlightGutterLine: { $def: false, set(n) { this.editor?.setOption('highlightGutterLine', n) } },
        highlightActiveLine: { $def: false, set(n) { this.editor?.setOption('highlightActiveLine', n) } },
        showCursor: true,
        isChanged: false
    },
    src: {
        $def: '',
        set(v) {
            this.value = v;
        },
    },

    get srcPatch() { 
        return ODA.rootPath + '/components/editors/ace-editor/src/'; 
    },
    get container() {
        return this.$?.('div');
    },
    async attached() {
        this.init();
    },
    init() {
        this.async(async () => {
            if (!this.container) return;
            this.editor = ace?.edit(this.container);
            this.editor.renderer.attachToShadowRoot();
            ['basePath', 'modePath', 'themePath', 'workerPath'].map(path => ace.config.set(path, this.srcPatch));
            await this.setTheme();
            await this.setMode();
            this.setOptions();
            this.setValue(this.src || this.value || '');
            this.src ||= this.editor.getValue();
            this.editor.setReadOnly(this.readOnly);
            this.editor.commands.addCommand({
                name: 'format',
                bindKey: { win: "Ctrl-Q", mac: "Option-q" },
                exec: async () => {
                    // https://github.com/beautify-web/js-beautify
                    await import('./src/beautify.js');
                    await import('./src/beautify-css.js');
                    await import('./src/beautify-html.js');
                    const mode = this.editor.session.getMode().$id,
                        fn = mode.includes('html') ? html_beautify : mode.includes('css') ? css_beautify : js_beautify,
                        session = this.editor.getSession();
                    session.setValue(fn(session.getValue(), { "end_with_newline": true, }));
                }
            })
            this.editor.session.on('change', (e) => {
                this.isChanged = this.editor.getValue() !== this.src;
                this.fire('change', this.editor.getValue());
            })
        })
    },
    setValue(value) {
        this.editor?.setValue(value);
        this.editor?.session.selection.clearSelection();
    },
    async setTheme(theme = this.theme || 'chrome') {
        await import(`${this.srcPatch}theme-${theme}.js`);
        this.editor?.setTheme(`ace/theme/${theme}`);
    },
    async setMode(mode = this.mode || 'javascript') {
        await import(`${this.srcPatch}mode-${mode}.js`);
        const _mode = ace.require(`ace/mode/${mode}`).Mode;
        this.editor?.session.setMode(new _mode());
    },
    setOptions(options = this.options || {}) {
        this.editor?.setOptions(options);
    },
    get options() {
        const options = { showPrintMargin: false };
        ['fontSize', 'maxLines', 'minLines', 'wrap', 'showGutter', 'highlightGutterLine', 'highlightActiveLine'].forEach(i => options[i] = this[i]);
        return options;
    }
})
