if (!globalThis.ROCKS) {
    const _addEventListener = EventTarget.prototype.addEventListener
    const _removeEventListener = EventTarget.prototype.removeEventListener

    Object.defineProperties(EventTarget.prototype, {
        addEventListener: {
            configurable: true,
            enumerable: true,
            writable: true,
            value: function addEventListener(event, handler, ...args) {
                if (!this.__events){
                    Object.defineProperty(this, '__events', {
                        value: new Map()
                    })
                }
                let array = this.__events.get(event);
                if (!array) {
                    array = [];
                    this.__events.set(event, array);
                }
                array.push({ handler, event: event });
                return _addEventListener.call(this, event, handler, ...args)
            }
        },
        removeEventListener: {
            configurable: true,
            enumerable: true,
            writable: true,
            value: function removeEventListener(event, handler, ...args) {
                if (this.__events && this.__events.has(event)) {
                    const events = this.__events.get(event);
                    const curEvent = events.find(i => i.handler === handler)
                    if (curEvent) {
                        const idx = events.indexOf(curEvent);
                        if (idx > -1) {
                            events.splice(idx, 1);
                        }
                    }
                    if (!events.length)
                        this.__events.delete(event);
                }
                return _removeEventListener.call(this, event, handler, ...args)
            }
        },
    })
    const OBS_PREFIX = '_obs__';
    globalThis.ROCKS = function (prototype){
        const rocks = class extends (this || EventTarget) {
            constructor() {
                super(...arguments);
                if (this.constructor.__rocks__.prototype.$observers){
                    this.async(()=>{
                        for(let obs in this.constructor.__rocks__.prototype.$observers){
                            this[OBS_PREFIX+obs];
                        }
                    })
                }
            }
            $core = {
                events: {},
                callbacks: new Map(),
                throttles: {},
                attributes: new Map(),
                debounces: new Map(),
            }
            async(handler, delay = 0) {
                if (typeof handler === 'string')
                    handler = this[handler].bind(this);
                const fn = delay? setTimeout : requestAnimationFrame; //todo requestAnimationFrame не будет работать в Node.js
                return fn(()=>{
                    handler();
                }, delay)
            }
            debounce(key, handler, delay = 0) {
                if (typeof handler === 'string')
                    handler = this[handler].bind(this);
                let db = this.$core.debounces.get(key);
                if (db)
                    delay ? clearTimeout(db) : cancelAnimationFrame(db); //todo cancelAnimationFrame не будет работать в Node.js
                const fn = delay ? setTimeout : requestAnimationFrame; //todo requestAnimationFrame не будет работать в Node.js
                const t = fn(() => {
                    this.$core.debounces.delete(key);
                    handler();
                }, delay);
                this.$core.debounces.set(key, t)
            }
            throttle(key, handler, delay = 0) {
                if (typeof handler === 'string')
                    handler = this[handler].bind(this);
                let task = this.$core.throttles[key];
                if (task) {
                    task.handler = handler;
                } else {
                    const fn = delay ? setTimeout : requestAnimationFrame; //todo requestAnimationFrame не будет работать в Node.js
                    const clearFn = delay ? clearTimeout : cancelAnimationFrame; //todo cancelAnimationFrame не будет работать в Node.js
                    task = {
                        handler,
                        id: fn(() => {
                            clearFn(task.id);
                            this.$core.throttles[key] = undefined;
                            task.handler();
                        }, delay)
                    };
                    this.$core.throttles[key] = task;
                }
            }
            listen(event, callback, props = { target: this, once: false, useCapture: false }) {

                if (typeof callback === 'string')
                    callback = this[callback];
                if (!callback) throw new Error (`Undefined event handler for "${event}"`);
                if(this.$core.callbacks.has(callback))
                    callback = this.$core.callbacks.get(callback)
                else
                    this.$core.callbacks.set(callback, (callback = callback.bind(this)))
                const handlers = this.$core.events[event] ??= [];
                if (handlers.includes(callback)) return;
                if (!props.once)
                    handlers.push(callback);
                (props.target || this).addEventListener(event, callback, props);
            }
            unlisten(event, callback, props = { target: this }){
                if (!this.$core.events[event]) return
                if (typeof callback === 'string')
                    callback = this[callback];
                callback = this.$core.callbacks.get(callback)
                if (callback){
                    const idx = this.$core.events[event].indexOf(callback)
                    if (~idx) this.$core.events[event].splice(idx, 1);
                    (props.target || this).removeEventListener(event, callback);
                }
                else{
                    while (callback = this.$core.events[event].pop()){
                        (props.target || this).removeEventListener(event, callback);
                    }
                }
            }
            fire(event, value){
                if (!this.__events?.has(event)) return;
                const ev = new CustomEvent(event, { detail: { value, target: this }, composed: true });
                this.dispatchEvent(ev);
            }
            async(handler, delay = 0) {
                if (typeof handler === 'string')
                    handler = this[handler].bind(this);
                const fn = delay?setTimeout:requestAnimationFrame;
                return fn(()=>{
                    handler();
                }, delay)
            }
            debounce(key, handler, delay = 0) {
                if (typeof handler === 'string')
                    handler = this[handler].bind(this);
                let db = this.$core.debounces.get(key);
                if (db)
                    delay ? clearTimeout(db) : cancelAnimationFrame(db);
                const fn = delay ? setTimeout : requestAnimationFrame;
                const t = fn(() => {
                    this.$core.debounces.delete(key);
                    handler();
                }, delay);
                this.$core.debounces.set(key, t)
            }
            throttle(key, handler, delay = 0) {
                if (typeof handler === 'string')
                    handler = this[handler].bind(this);
                let task = this.$core.throttles[key];
                if (task) {
                    task.handler = handler;
                } else {
                    const fn = delay ? setTimeout : requestAnimationFrame;
                    const clearFn = delay ? clearTimeout : cancelAnimationFrame;
                    task = {
                        handler,
                        id: fn(() => {
                            clearFn(task.id);
                            this.$core.throttles[key] = undefined;
                            task.handler();
                        }, delay)
                    };
                    this.$core.throttles[key] = task;
                }
            }
            $super(name, ...args) {
                //                  rocks     class     rocks     class
                let proto = super.__proto__.__proto__.__proto__.__proto__;
                let ds;
                while (!ds && proto) {
                    ds = Object.getOwnPropertyDescriptor(proto, name);
                    //              rocks     class
                    proto = proto.__proto__.__proto__;
                }
                if(ds.set?.setter && args.length)
                    return ds.set.setter.call(this, ...args);
                
                if (ds.get?.getter)
                    return ds.get.getter.call(this);
                    
                if (typeof ds.value === "function")
                    return ds.value.call(this, ...args);
            }
            static get name() {
                let parent = this.prototype.__proto__.constructor
                let name = parent.name;
                while (name === 'rocks') {
                    parent = parent.prototype.__proto__.constructor
                    name = parent.name
                }
                return name;
            }
        }
        const descrs = {}
        if (prototype.extends) {
            if (typeof prototype.extends === 'string') {
                prototype.extends = str2arr(prototype.extends);
            }
        } else {
            prototype.extends = [];
        }
        prototype.extends.forEach(ext => {
            joinProps.call(this,  descrs, prototype, ext.__rocks__);
        })
        joinProps.call(this, descrs, prototype, this?.__rocks__)
        if (prototype.$observers) {
            for (const name in prototype.$observers) {
                let func = prototype.$observers[name]
                let expr;
                let args;

                if (Array.isArray(func) || typeof func === 'string') {
                    if (typeof func === 'string') {
                        args = str2arr(func);
                    }
                    func = prototype[name];
                }
                if (!args) {
                    expr = func.toString();
                    const argsStart = expr.indexOf('(') + 1;
                    const argsEnd = expr.indexOf(')', argsStart);
                    args = str2arr(expr.slice(argsStart, argsEnd));
                }

                const checkArgs = args.map(a => a.split('.')[0].replace('?', ''));
                const obsName = `${OBS_PREFIX}${name}`
                prototype[obsName] = {
                    get() {
                        checkArgs.forEach(a => {
                            if (!(a in this)) {
                                let message = `undefined property "${a}" in ${prototype.is ? `<${prototype.is}>` : this.constructor.name} for observer ${name}`;
                                let host = this.domHost;
                                while (host){
                                    if (a in host){
                                        message += `
But this problem can be fixed by set
the property modifier $pdp: true
in the <${host.localName}>`;
                                        break;
                                    }
                                    host = host.domHost;
                                }
                                throw new Error(message);
                            }
                        })
                        const props = args.map(a => this[a]);
                        if (props.every(p => p !== undefined)) {
                            this.debounce(obsName, () => {
                                this[name](...props);
                            })
                        }
                    }
                };
                if (func) prototype[name] = func;
            }
        }

        joinProps.call(this,  descrs, prototype, {descrs: convertor(prototype)})
        Object.defineProperty(rocks, '__rocks__', {configurable: false, value: {descrs, prototype}})
        try{
            Object.defineProperties(rocks.prototype, rocks.__rocks__.descrs);
        }
        catch (e){
            switch (e.name){
                case 'TypeError':{
                    const err = Object.values(rocks.__rocks__.descrs).find(i=>{
                        return i.value && (i.get || i.set || i.writable)
                    })
                    if (err)
                        throw new Error(e.name +' on "' + err.name+'" property: '+e.message)
                    else
                        throw e;
                } break;
            }
        }
        return rocks;
    }
    const ROCKS = globalThis.ROCKS;
    Object.defineProperty(Object.__proto__, 'ROCKS', {enumerable: false, configurable: false, value: ROCKS})
    function joinProps(descrs, proto, source){
        for (let d in source?.descrs){
            if (!descrs[d] || descrs[d].configurable) {
                descrs[d] ??= Object.create(null)
                Object.assign(descrs[d], source.descrs[d]);
                continue;
            }
            throw new Error(`Cannot override protected property "${d}"`); //todo указывать первоисточник где protected
        }
        for (let p of PROTOTYPE_SPECIAL_GROUPS)
            proto[p] = Object.assign(proto[p] || {}, source?.prototype?.[p] || {})
    }
    function str2arr(str) {
        return str.split(',').map(s => s.trim()).filter(Boolean);
    }
    const KEY = Symbol();
    ROCKS.KEY = KEY;
    function reactor(target) {
        if (!Object.isExtensible(target) || (target.constructor !== Object &&  target.constructor !== Array) || target?.constructor === Promise)
            return target;

        let op = target[KEY];
        if (op){
            op.hosts.add(this);
            return op.proxy;
        }
        op = Object.create(null);
        const handlers = {
            get: (target, key, resolver) => {
                let val = op.target[key];
                if (val){
                    if (op === val || op.target === val || key.constructor === Symbol || (val instanceof Object && val.constructor.prototype === val) || val instanceof Function)
                        return val;
                }
                const $prop = op.props[key] ??= {deps: new Set()}
                $prop.deps.add(DEP_TARGET || target);
                return reactor.call(this, val);
            },
            set: (target, key, value) => {
                const old = (Array.isArray(op.target) && key === 'length')?undefined:op.target[key];
                if (Object.equal(old, value)) return true;
                const $prop = op.props[key] ??= {deps: new Set()}
                op.target[key] = reactor.call(this, value);
                resetDeps($prop);
                return true;
            }
        };
        const proxy = new Proxy(op?.target || target, handlers);
        if (!target[KEY]) {
            op.target = target;
            op.hosts = new Set();
            op.proxy = proxy;
            op.props = Object.create(null);
            Object.defineProperty(target, KEY, {
                enumerable: false,
                configurable: false,
                value: op
            });
        }
        op.hosts.add(this);
        return proxy;
    }
    Object.equal ??= function (a, b, recurse) {
        if (a === b) return true;
        if (!(a instanceof Object) || !(b instanceof Object))
            return false;
        if ((a?.[KEY] || a) === (b?.[KEY] || b))
            return true;
        if (a instanceof Function && a.constructor === b.constructor)
            return a.toString() === b.toString();
        if (recurse) {
            for (let key in Object.assign({}, a, b))
                if (!Object.equal(b[key], a[key], recurse)) return false;
            return true;
        }
        return false;
    };
    function isObject(obj) {
        return obj && typeof obj === 'object';
    }
    function isNativeObject(obj) {
        return obj && (obj.constructor === Object);
    }
    function convertor(proto, decor, result){
        const descriptors = Object.getOwnPropertyDescriptors(proto);
        result ??= Object.create(null);
        for (let name in descriptors){
            if (PROTOTYPE_RESERVED_WORDS.includes(name))
                continue;

            if (decor && PROTOTYPE_SPECIAL_GROUPS.includes(name))
                continue;
            if (decor && PROPERTY_ATTRIBUTES.includes(name))
                continue;
            const d = descriptors[name];
            let prop = Object.create(null);

            switch (typeof d.value){
                case 'object':{
                    if(d.value === null){
                        prop.$type = Object;
                        prop.$def = null;
                    }
                    else if (Array.isArray(d.value)){
                        prop.$type = Array;
                        prop.$def = d.value;
                    }
                    else{
                        const attrs = Object.keys(d.value);
                        if (attrs.some(i=>PROPERTY_SIGNS.includes(i))) { // it is property
                            for (let attr of attrs){
                                if (!PROPERTY_ATTRIBUTES.includes(attr) && !PROPERTY_SIGNS.includes(attr)){
                                    throw new Error(`Unknown attribute "${attr}" in description for property "${name}".`)
                                }
                                prop[attr] ??= d.value[attr];
                            }
                            if (!prop.$type && prop.$def !== undefined){
                                switch (typeof prop.$def){
                                    case 'object':{
                                        if (Array.isArray(prop.$def))
                                            prop.$type = Array;
                                        else if(prop.$def === null)
                                            prop.$type = Object;
                                        else
                                            prop.$type = prop.$def.__proto__?.constructor || Object;
                                    } break;
                                    default:{
                                        prop.$type = prop.$def.__proto__.constructor;
                                    }
                                }
                            }
                        }
                        else if (ROCKS.PROPERTY_FLAGS.includes(name)){
                            for (let attr of attrs){
                                if (ROCKS.PROPERTY_FLAGS.includes(attr)){
                                    prop[attr] ??= d.value[attr];
                                }
                            }
                            prop[name] = true;
                            convertor(d.value, prop, result);
                            continue;
                        }
                        else if (attrs.some(i => PROPERTY_ATTRIBUTES.includes(i))) {  // it is group
                            for (let attr of attrs){
                                if (PROPERTY_ATTRIBUTES.includes(attr)){
                                    prop[attr] ??= d.value[attr];
                                }
                            }
                            prop.$group ??= name;
                            convertor(d.value, prop, result);
                            continue;
                        }
                        else{
                            prop.$type = d.value.__proto__?.constructor || Object;
                            prop.$def = d.value;
                        }
                    }
                } break;
                case 'function':{
                    if (PROPERTY_TYPES.includes(d.value.name))
                        prop.$type = globalThis[d.value.name];
                    else
                        prop.value = d.value;
                } break;
                case 'undefined':{
                    if (d.get)
                        prop.get = d.get;
                    if (d.set)
                        prop.set = d.set;
                } break;
                default:{
                    prop.$type = d.value.__proto__.constructor;
                    prop.$def = d.value;
                }
            }
            if (name in result)
                throw new Error(`${prop.value?'Method':'Property'} "${name}" redeclared.`);

            prop = Object.assign(prop, decor || {})
            prop.name = name;
            if (name.startsWith(OBS_PREFIX)){
                prop.isObserver = true;
            }
            if (prop.$attr){
                if (prop.$attr === true)
                    prop.$attr = name.toKebabCase();
            }
            if (!prop.value){
                if (prop.$def !== undefined){
                    const def = prop.$def;
                    switch (prop.$type) {
                        case Object: {
                            prop.$def = function () { return def?Object.assign({}, def):def};
                        } break;
                        case Array: {
                            prop.$def = function () { return Array.from(def)};
                        } break;
                        default:{
                            prop.$def = function () { return def};
                        }
                    }
                }
                prop.toType = getTypeConverter(prop.$type);
                prop.event = name.toKebabCase()+'-changed';
                const key = '#'+name;
                if (prop.get)
                    prop.getter = prop.get;
                prop.get = function (){
                    let  $prop = joinPropDescriptors.call(this, name);
                    let val = this[key];
                    if (val === undefined){
                        if($prop.getter){
                            $prop.depTarget = DEP_TARGET;
                            DEP_TARGET = $prop;
                            val = $prop.getter.call(this);
                            DEP_TARGET = $prop.depTarget;
                        }
                        if($prop.$save && this.$loadPropValue)
                            val = this.$loadPropValue(name);

                        if ($prop.$def && !(key in this) && (val === undefined || Number.isNaN(val))) {
                            val = $prop.$def();
                            if (!$prop.$freeze)
                                val = reactor.call(this, val);
                            $prop.old = this[key] = val;
                            this.$notify?.($prop, val);
                        }
                        if (Object.equal(val, $prop.old, true)){
                            this[key] = $prop.old;
                        }
                        else if (!(val instanceof Promise)){
                            if (!$prop.$freeze)
                                val = reactor.call(this, val);
                            $prop.old = this[key] = val;
                            resetDeps($prop);
                            if($prop.$public)
                            this.fire($prop.event, val);
                            this.$notify?.($prop, val);
                        }
                        else
                            $prop.old = this[key] = val;

                        val?.length; //todo никогда не удалять
                    }
                    if (!$prop.isObserver)
                        $prop.deps.add(DEP_TARGET || this);
                    return $prop.toType(val);
                }
                // нужно в $supper
                prop.get.getter = prop.getter;
                if (prop.set)
                    prop.setter = prop.set;
                prop.set = function (val){
                    let $prop = joinPropDescriptors.call(this, name);
                    if ($prop.$readOnly)
                        throw new Error('Read only!!! ' + name);
                    val = $prop.toType(val);
                    $prop.old = this[key];
                    if (Object.equal(val, $prop.old, true)) return;
                    if (!$prop.$freeze)
                        val = reactor.call(this, val);
                    this[key] = val;
                    $prop.setter?.call(this, val, $prop.old);
                    if($prop.$save && this.$savePropValue)
                        this.$savePropValue(name, val);
                    resetDeps($prop);
                    this.fire($prop.event, val);
                    this.$notify?.($prop, val);
                }
                // нужно в $supper
                prop.set.setter = prop.setter;
            }
            else{
                if (name.startsWith('__'))
                    prop.$final = true;
                else if (!name.startsWith('_')){
                    prop.$public = true;
                    prop.$pdp = true;
                }
            }
            prop.enumerable = prop.$public || prop.$pdp || false;
            prop.configurable = !prop.$final;
            result[name] = prop;
        }
        return result;
    }
    function joinPropDescriptors(name){
        const prop = this.constructor.__rocks__.descrs[name];
        prop.hosts ??= new WeakMap();
        let res = prop.hosts.get(this);
        if (!res){
            res = Object.assign({host: this, key: '#'+name, deps: new Set()}, prop);
            prop.hosts.set(this, res);
        }
        return  res;
    }
    function resetDeps($prop, stack = new WeakSet) {
        for (let p of $prop.deps || []){
            if (p === $prop || p.$freeze)
                continue;
            if (p.host) // todo опасно привязываться к хосту
                p.host[p.key] = undefined;
            if (stack.has(p)) continue;
            stack.add(p);
            resetDeps(p, stack);
        }
        if ($prop.isObserver){
            $prop.host.throttle($prop.name, ()=>{
                $prop.host[$prop.name];
            }, 16)
        }
        const h = $prop.host || $prop;
        if (h?.$render){
            h.throttle('$render', ()=>{
                h.$render?.();
            }, 16)
        }
        else{
            for (let h of $prop[KEY]?.hosts || []){
                if (h?.$render){
                    h.throttle('$render', ()=>{
                        h.$render?.();
                    }, 16)
                }
            }
        }
    }
    function toDate(v){return new Date(v)}
    function toString(v){return v?.toString() || ''}
    function toNumber(v){return parseFloat(v) || 0}
    function toBigInt(v){return BigInt(v) || 0n}
    function toBool(v, def = false) {
        if (v === undefined || v === null)
            return def;
        switch (typeof v) {
            case 'object': return true;
            case 'string': return v.toLowerCase() === 'true';
            case 'boolean': return v;
            case 'number': return v !== 0;
        }
        return false;
    }
    function getTypeConverter(type){
        switch (type) {
            case Boolean: return toBool;
            case Number: return toNumber;
            case String: return toString;
            case Date: return toDate;
            case BigInt: return toBigInt;
        }
        return (val)=>{
            return val;
        }
    }
    let DEP_TARGET;
    const PROTOTYPE_RESERVED_WORDS = ['is', 'imports', 'extends'];
    ROCKS.PROPERTY_FLAGS = ['$save', '$multiSelect', '$final', '$public', '$pdp', '$readOnly', '$freeze'];
    const PROTOTYPE_SPECIAL_GROUPS = ['$listeners', '$observers', '$keyBindings'];
    const PROPERTY_TYPES = ['Boolean', 'Number', 'String', 'Array', 'Object', 'Date', 'BigInt'];
    const PROPERTY_SIGNS = ['$def', '$type', 'get', 'set'];
    const PROPERTY_ATTRIBUTES = [...ROCKS.PROPERTY_FLAGS, '$label', '$list', '$attr', '$editor', '$group', '$description'];
    Array:{
        Object.defineProperty(Array.prototype, 'has', {
            enumerable: false, configurable: true, value: Array.prototype.includes
        });
        Object.defineProperty(Array.prototype, 'clear', {
            enumerable: false, configurable: true, value: function () {
                this.splice(0);
            }
        });
        Object.defineProperty(Array.prototype, 'last', {
            enumerable: false, configurable: true, get() {
                return this[this.length - 1];
            }
        });
        Object.defineProperty(Array.prototype, 'add', {
            enumerable: false, configurable: true, value: function (...item) {
                let index = -1;
                for (let i of item) {
                    index = this.indexOf(i);
                    if (index>-1) continue;
                    index = this.push(i);
                    index--;
                }
                return index;
            }
        });
        Object.defineProperty(Array.prototype, 'remove', {
            enumerable: false, configurable: true, value: function (...items) {
                for (const item of items) {
                    let idx = this.indexOf(item);
                    if (~idx)
                        this.splice(idx, 1);
                }
            }
        });
    }
    String:{
        const kebabGlossary = Object.create(null);
        function toKebab(str) {
            return kebabGlossary[str] ??= str.replace(/\B([A-Z])/g, '-$1').toLowerCase();
        }
        if (!String.toKebabCase) {
            Object.defineProperty(String.prototype, 'toKebabCase', {
                enumerable: false, value: function () {
                    return toKebab(this.toString());
                }
            });
        }
        const camelGlossary = Object.create(null);
        function toCamel(str) {
            return camelGlossary[str] ??= str.replace(/-(\w)/g, function (_, c) { return c ? c.toUpperCase() : '' })
        }

        if (!String.toCamelCase) {
            Object.defineProperty(String.prototype, 'toCamelCase', {
                enumerable: false, value: function () {
                    return toCamel(this.toString());
                }
            });
        }

        const capitalGlossary = Object.create(null);
        function toCapital(str) {
            if (!str) return '';
            return capitalGlossary[str] ??= str[0].toUpperCase() + str.slice(1);
        }

        if (!String.toCapitalCase) {
            Object.defineProperty(String.prototype, 'toCapitalCase', {
                enumerable: false, value: function () {
                    return toCapital(this.toString());
                }
            });
        }

        if (!String.toQName) {
            Object.defineProperty(String.prototype, 'toQName', {
                enumerable: false, value: function () {
                    return this.toLowerCase().split(' ')
                    .map((s, i) => {
                        if (i === 0) return (s === 'the') ? '' : s;
                        if (s.length < 5) return s;
                        return s.substring(0, 5);
                    })
                    .join('-')
                    .replace(/-{2,}/g, '-')
                    .replace(/(^\d)/, '_$1')
                    .replace(/\./g, '');
                }
            })
        }
    }
}
export default globalThis.ROCKS;