ODA({is: 'oda-property-grid', imports: '@oda/table', extends: 'this, oda-table',
    template: /*html*/`
        <style>
            :host {
                max-height: 100%;
                border: 1px solid gray;
            }
        </style>
        <slot ~show="!inspectedObject" @slotchange="onSlot"></slot>
    `,
    get columns() {
        return [
            { name: 'name', cellTemplate: 'oda-pg-cell-name', headerTemplate: 'oda-property-grid-header-cell-name', label: 'Property', treeMode: true, fix: 'left', $sort: 1 },
            { name: 'value', cellTemplate: 'oda-pg-cell-value', headerTemplate: 'oda-property-grid-header-cell-value' },
            { name: 'category', $hidden: true, $sortGroups: 0 },
        ]
    },
    get groups(){
        return this.columns.filter(c => c.name === 'category');
    },
    get label() {
        if (!this.inspectedObject) return;
        if (Array.isArray(this.inspectedObject)) {
            if (this.inspectedObject.length > 1)
                return this.inspectedObject[0]?.constructor?.name + ` [${this.inspectedObject.length}]`
            return this.inspectedObject[0]?.constructor?.name;
        }
        return this.inspectedObject?.constructor?.name
    },
    $final:{
        $readOnly: true,
        lazy: true,
        autoWidth: true,
        allowFocus: true,
        showHeader: true,
        showFooter: true,
        colLines: true,
        rowLines: true,
    },

    $public: {
        $pdp: true,
        autoFixRows: true,
        inspectedObject: Object,
        icon: 'icons:settings',
        expertMode: false,
        filterByFlags:{
            $def: '',
            get $list(){
                return ROCKS.PROPERTY_FLAGS
            },
            $multiSelect: true
        },
        groupExpandingMode: 'first',
    },
    get dataSet() {
        if (this.inspectedObject){
            const ds = this.PropertyGridDataSet;
            ds.$expanded = true;
            return ds.items;

        }
        return [];
    },
    get PropertyGridDataSet() {
        return new PropertyGridDataSet(this.inspectedObject, this.expertMode)
    },
    onSlot(e) {
        this.inspectedObject ??= e.target.assignedElements();
    },
    _sort(array = []) {
        if (!this.sorts.length) return;
        array.sort((a, b) => {
            if (!a.prop !== !b.prop) {
                return a.prop ? -1 : 1;
            }
            let res = 0;
            this.sorts.some(col => {
                const _a = a[col[this.columnId]];
                const _b = b[col[this.columnId]];
                if (!isNaN(_a) && !isNaN(_b)) {
                    res = parseFloat(_a) > parseFloat(_b) ? 1 : -1
                } else {
                    res = (String(_a)).localeCompare(String(_b)) * col.$sort;
                }
                if (res) return true;
            });
            return res;
        });
    },
})

class PropertyGridDataRowOwner extends ROCKS({
    $expanded:{
        $def: false,
        get(){
            return false
        }
    },
    $hasChildren: undefined,
    get items() {
        if (this.mixed)
            return [];
        if (!this.$expanded){
            if (this.value && typeof this.value === 'object'){
                if (this.value.props){
                    if (Object.keys(this.value.props).length)
                        return [{}];
                }
                else if (Object.keys(this.value).length){
                    return [{}];
                }

            }
            return [];
        }
        const items = {}
        for (let obj of this.inspectedObjects || []) {
            if (!(obj instanceof Object)) continue;
            const isRocks = obj.constructor.__rocks__;
            let proto = obj;
            while (proto) {
                const descriptors = proto.constructor.__rocks__?.descrs || Object.getOwnPropertyDescriptors(proto);
                for (let name in descriptors) {
                    const d = descriptors[name];
                    if (d.value?.constructor === Function)
                        continue;
                    if (!this.dataSet.expert){
                        if (!d.enumerable)
                            continue;
                    }
                    const row = items[name] ??= new PropertyGridDataRow(name, this.dataSet, proto, d);
                    row.inspectedObjects?.add(obj);
                }
                if (!this.dataSet.expert)
                    break;
                proto = proto.__proto__;
            }
        }
        return Object.values(items);
    }
}){}


class MixedArray extends Array {

}
class PropertyGridDataRow extends PropertyGridDataRowOwner.ROCKS({
    get io(){
        return this.dataSet.inspectedObjects[0];
    },
    get category() {
        let cat = this.prototype?.toString() || '';
        if (this.prop?.$group)
            cat += ': ' + this.prop?.$group;
        return cat
    },
    get mixed() {
        return this.value instanceof MixedArray;
    },
    _extractEditor(s) { // @path/path[component-name]
        const parts = s.split('[');
        if (parts.length === 2){
            return { path: parts[0], tag: parts[1].slice(0, -1) };
        }
        return { path: undefined, tag: s };
    },
    _getDefaultEditor() {
        switch (this.prop?.$type || typeof this.value) {
            case Array:
            case Object:
            case 'object':
                return 'oda-pg-object';
            case Number:
            case 'number':
                return 'oda-pg-number';
            case Boolean:
            case 'boolean':
                return 'oda-pg-bool';
            case String:
            case 'string':
            default:
                return 'oda-pg-string';
        }
    },
    get editor() {
        if (this.mixed)
            return 'oda-pg-mixed';
        const editor =  this.prop?.$editor || this.value?.$editor;
        if (editor){
            const onError = (err) => {
                console.error(`Type editor to "${this.name}" prop not loaded.\n`, err);
                this.editor = this._getDefaultEditor();
            }
            try {
                const { path, tag } = this._extractEditor(editor);
                if (path){
                    let url = this.inspectedObjects[0]?.url || '';
                    Promise.resolve(ODA.import((url ? (url + '/~/') : '') + path)).catch(onError);
                }
                return tag;
            } catch (err) {
                onError(err);
            }
        }
        return this._getDefaultEditor();
    },
    get value() {
        const list = this.inspectedObjects.map(io => io[this.name])
        const value = list[0];
        if (typeof value !== 'object' && list.some(v => v !== value))
            return new MixedArray(...list);
        return value;
    },
    set value(n) {
        if (this.ro) return;
        for (let io of this.inspectedObjects) {
            if(io[this.name] !== n)
                io[this.name] = n;
        }
    },
    get ro() {
        if (this.prop){
            if (this.prop?.$readOnly)
                return true;
            if (!this.prop.set)
                return true;
            //todo возможно надо обработать value не Function
            return false;
        }
        return true;
    },
    get list(){
        return this.prop?.$list;
    },
    get defaultValue(){
        if (typeof this.prop?.$def === 'function')
            return this.prop.$def();
        return this.prop?.$def;
    },
    $expanded:{
        $def: false,
        set(n){
            if (this.value?.then) {
                this.value?.then(obj => {
                    const row = new PropertyGridDataRow('Resolve', this.dataSet)
                    row.isRevoked = true;
                    row.value = obj;
                    row.ro = true;
                    if (typeof obj === 'object')
                        row.inspectedObjects = [obj];
                    this.items = [row];
                }).catch(e => {
                    const row = new PropertyGridDataRow('Reject', this.dataSet)
                    row.isRevoked = true;
                    row.value = e;
                    row.ro = true;
                    row.inspectedObjects = [e];
                    this.items = [row];
                })
            }
        }
    }
}){
    constructor(name, dataSet, prototype, prop) {
        super(name, dataSet, prototype)
        this.dataSet = dataSet;
        this.prototype = prototype;
        this.prop = prop;
        this.name = name;
        this.inspectedObjects = [];
    }
}

class PropertyGridDataSet extends PropertyGridDataRowOwner.ROCKS({
    get dataSet() {
        return this;
    },
    isRoot: true,

}){
    constructor(inspectedObject, expert) {
        super(inspectedObject, expert);
        this.expert = expert;
        if (!Array.isArray(inspectedObject))
            this.inspectedObjects = [inspectedObject];
        else if (inspectedObject.every(i => typeof i === 'object'))
            this.inspectedObjects = inspectedObject;
        else
            this.inspectedObjects = [inspectedObject];
    }
}

cells: {
    ODA({is: 'oda-pg-cell-value', extends: 'oda-table-cell',
        template: /*html*/`
        <style>
            :host {
                overflow: hidden;
                @apply --horizontal;
            }
            :host > span {
                @apply --dimmed;
                user-select: text;
            }
            .editor {
                border: none !important;
            }
        </style>
        <oda-button ~if="list?.length" @tap.stop.prevent="showDD" icon="icons:chevron-right:90"></oda-button>
        `,
        $public: {
            get template() {
                return this.item?.editor;
            },
            get list() {
                if (this.item?.list?.then) {
                    this.item.list.then(list => {
                        this.list = list;
                    })
                    return undefined;
                }
                return this.item?.list;
            },
        },
        async showDD(e) {
            if (this.__dd_control){
                this.__dd_control.fire('cancel');
            }
            else{
                this.__dd_control = ODA.createElement('oda-menu', { items: this.list.map(i => ({ label: i?.label ?? i?.name ?? i, value: i.value || i })), selectedItem: this.item.value });
                try {
                    const res = await ODA.showDropdown(this.__dd_control, {}, { useParentWidth: true, parent: e.target.domHost, fadein: true});
                    this.item.value = res.focusedItem.value;
                }
                catch (e) {

                }
            }

            delete this.__dd_control;
            this.__dd_control = undefined;
        },
        resetValue() {
            this.item.value = this.item.default;
        }
    })
    ODA({is: 'oda-pg-cell-name', extends: 'oda-table-cell',
        template: /*html*/`
        <style>
            :host {
                font-weight: {{(item?.prop)?'bold':'normal'}};
            }
        </style>
        <oda-button :title="item.defaultValue" ~if="item.defaultValue" :disabled @tap.stop.prevent="resetValue" icon="av:replay"></oda-button>
    `,
        title: {
            get() {
                return this.item.name;
            },
            $attr: true,
        },
        get disabled(){
            return Object.equal(this.item.defaultValue, this.item.value);
        },
        resetValue() {
            this.item.value = this.item.defaultValue;
        },
        get value() {
            return this.item.label || this.item.name;
        }
    })

    ODA({is: 'oda-property-grid-header-cell-name', extends:'oda-table-header-cell',
        template: /*html*/`
            <style>
                :host {
                    @apply --horizontal;
                    align-items: center;
                }
            </style>
            <oda-button ~if="allow" class="no-flex"  @tap.stop.prevent="resetValue" icon="av:replay" slot="tools" title="Reset all defaults"></oda-button>
        `,
        get allow(){
            return this.items.filter(i=>{
                return i.defaultValue !== undefined && i.defaultValue !== i.value;
            }).length > 0;
        },
        resetValue() {
            let io = this.inspectedObject
            if (!Array.isArray(io)){
                io = [io]
            }
            io.forEach(i=>{
                i.$resetSettings();
            })
            this.items.forEach(item => {
                if (item.prop?.default === undefined) return;
                item.value = (typeof item.prop.default === 'function') ? item.prop.default() : item.prop.default;
            })
        }
    })
    ODA({is: 'oda-property-grid-header-cell-value',
        template: /*html*/`
            <style>
                :host {
                    @apply --horizontal;
                    align-items: center;
                    font-size: small;
                }
            </style>
            <div style="padding: 4px;" class="flex horizontal">
                <label style="text-align: center;" class="flex">Value</label>
            </div>
            <oda-button class="no-flex" allow-toggle ::toggled="expertMode" icon="social:school"></oda-button>
        `
    })
}
editors: {
    ODA({is: 'oda-pg-object',
        template: /*html*/`
        <style>
            :host {
                @apply --disabled;
            }
        </style>
        {{text}}
        `,
        value: null,
        get text() {
            if (!this.value)
                return '[Object: undefined]';
            if (Array.isArray(this.value))
                return `[Array (${this.value.length})]`
            if (typeof this.value === 'object')
                return '[' + (this.value?.constructor?.name || typeof this.value) + ']';
            return this.value;
        }
    })

    ODA({is: 'oda-pg-mixed',
        template: /*html*/`
        <input :placeholder="'mixed: [' + item.value+']'" class="error flex content" type="text" style="border: none; outline: none; min-width: 0;width: 100%;" :readonly="item.ro === true"  @input="_input" @keydown.stop>
        `,
        _input(e) {
            this.item.value = e.target.value;
        }
    })

    ODA({is: 'oda-pg-string',
        template: /*html*/`
        <style>
            :host > input {
                font-size: medium;
                overflow: hidden;
                text-overflow: ellipsis;
            }
            :host > input[readonly] {
                @apply --dimmed;
            }
        </style>
        <input class="flex content" type="text" style="border: none; outline: none; min-width: 0;width: 100%;" ::value="item.value" :readonly="item.ro === true" @keydown.stop>
        `,
    })
    ODA({is: 'oda-pg-number',
        template: /*html*/`
        <style>
            :host > input[readonly] {
                @apply --dimmed;
            }
        </style>
        <input class="flex content"  style="border: none; outline: none; min-width: 0;width: 100%;"  type="number" ::value :readonly="item.ro === true" @keydown.stop>
        `,
        get value() {
            return +this.item.value;
        },
        set value(n) {
            this.item.value = +n;
        }
    })
    ODA({is: 'oda-pg-bool', imports: '@oda/checkbox',
        template: /*html*/`
            <style>
                :host {
                    @apply --horizontal;
                    @apply --flex
                    align-items: center;
                }
            </style>
            <oda-checkbox class="flex" ::value="item.value" style="justify-content: center;" :readonly="item.ro === true"></oda-checkbox>
        `,
    })
}
